#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Sep 30 11:48:05 2019

@author: zandieh
"""
import numpy as np
import torch
import time
import random
from numpy import linalg as npLA
from TensorSketchCuda import TensorSketch
from PolySketch import OblvPolySFeat , OblvPolySNTKFeat
from CorrelatedPolySketch import CorrelatedTensorSketch , OblvCorrPolySFeat, OblvCorrPolySNTKFeat
from RowNormSampling import RNSampler , NTK_RNSampler
from LeverageScore import RLSSampler , NTK_RLSSampler
from FourierFeatures import FF
from NystromFeatures import NF
import math



def PolySketch_KRR(s, q, m, lambda_, device_, problem_type, Xtrain, Ytrain, Xtest, Ytest):
    start_time = time.time()
    n = Xtrain.shape[0]
    d = Xtest.shape[1]
    
    Tensor_Sketch = TensorSketch(d, m, q, False, device_)
    
    
    Features_train = OblvPolySFeat(s, Tensor_Sketch, torch.from_numpy(Xtrain).to(Tensor_Sketch.device_) ).to('cpu').data.numpy()
    Features_test = OblvPolySFeat(s, Tensor_Sketch, torch.from_numpy(Xtest).to(Tensor_Sketch.device_) ).to('cpu').data.numpy()
    
    
    w_star = npLA.solve(np.dot(Features_train.T, Features_train)+lambda_*np.eye(Features_train.shape[1]), np.dot(Features_train.T, Ytrain))
    
    est = np.dot(Features_test, w_star)
    
    if problem_type == 'regress':
        
        TestErr = npLA.norm(est - Ytest)/np.sqrt(len(Ytest))
        
    elif problem_type == 'class':
        
        est = (est == est.max(axis=1, keepdims=True)).astype(int) - 1/Ytest.shape[1]
        TestErr = 100 - 100*npLA.norm(np.count_nonzero(est - Ytest, axis=1), 0)/len(Ytest)

    
    
    elapsed_time = time.time() - start_time
    
    return TestErr, elapsed_time




def PolySketch_NTK_KRR(L, q, m, lambda_, device_, problem_type, Xtrain, Ytrain, Xtest, Ytest):
    start_time = time.time()
    n = Xtrain.shape[0]
    d = Xtest.shape[1]
    
    Tensor_Sketch = TensorSketch(d, m, q, False, device_)
    
    
    Features_train = OblvPolySNTKFeat(L, Tensor_Sketch, torch.from_numpy(Xtrain).to(Tensor_Sketch.device_) ).to('cpu').data.numpy()
    Features_test = OblvPolySNTKFeat(L, Tensor_Sketch, torch.from_numpy(Xtest).to(Tensor_Sketch.device_) ).to('cpu').data.numpy()
    
    
    w_star = npLA.solve(np.dot(Features_train.T, Features_train)+lambda_*np.eye(Features_train.shape[1]), np.dot(Features_train.T, Ytrain))
    
    est = np.dot(Features_test, w_star)
    
    if problem_type == 'regress':
        
        TestErr = npLA.norm(est - Ytest)/np.sqrt(len(Ytest))
        
    elif problem_type == 'class':
        
        est = (est == est.max(axis=1, keepdims=True)).astype(int) - 1/Ytest.shape[1]
        TestErr = 100 - 100*npLA.norm(np.count_nonzero(est - Ytest, axis=1), 0)/len(Ytest)

    
    
    elapsed_time = time.time() - start_time
    
    return TestErr, elapsed_time



def CorrPolySketch_KRR(s, q, m, lambda_, device_, problem_type, Xtrain, Ytrain, Xtest, Ytest):
    start_time = time.time()
    n = Xtrain.shape[0]
    d = Xtest.shape[1]
    
    CorrTensor_Sketch = CorrelatedTensorSketch(d, m, q, device_)
    
    
    Features_train = OblvCorrPolySFeat(s, CorrTensor_Sketch, torch.from_numpy(Xtrain).to(CorrTensor_Sketch.device_) ).to('cpu').data.numpy()
    Features_test = OblvCorrPolySFeat(s, CorrTensor_Sketch, torch.from_numpy(Xtest).to(CorrTensor_Sketch.device_) ).to('cpu').data.numpy()
    
    
    w_star = npLA.solve(np.dot(Features_train.T, Features_train)+lambda_*np.eye(Features_train.shape[1]), np.dot(Features_train.T, Ytrain))
    
    est = np.dot(Features_test, w_star)
    
        
    if problem_type == 'regress':
        
        TestErr = npLA.norm(est - Ytest)/np.sqrt(len(Ytest))
        
    elif problem_type == 'class':
        
        est = (est == est.max(axis=1, keepdims=True)).astype(int) - 1/Ytest.shape[1]
        TestErr = 100 - 100*npLA.norm(np.count_nonzero(est - Ytest, axis=1), 0)/len(Ytest)
    
    
    elapsed_time = time.time() - start_time
    
    return TestErr, elapsed_time



def CorrPolySketch_NTK_KRR(L, q, m, lambda_, device_, problem_type, Xtrain, Ytrain, Xtest, Ytest):
    start_time = time.time()
    n = Xtrain.shape[0]
    d = Xtest.shape[1]
    
    CorrTensor_Sketch = CorrelatedTensorSketch(d, m, q, device_)
    
    
    Features_train = OblvCorrPolySNTKFeat(L, CorrTensor_Sketch, torch.from_numpy(Xtrain).to(CorrTensor_Sketch.device_) ).to('cpu').data.numpy()
    Features_test = OblvCorrPolySNTKFeat(L, CorrTensor_Sketch, torch.from_numpy(Xtest).to(CorrTensor_Sketch.device_) ).to('cpu').data.numpy()
    
    
    w_star = npLA.solve(np.dot(Features_train.T, Features_train)+lambda_*np.eye(Features_train.shape[1]), np.dot(Features_train.T, Ytrain))
    
    est = np.dot(Features_test, w_star)
    
        
    if problem_type == 'regress':
        
        TestErr = npLA.norm(est - Ytest)/np.sqrt(len(Ytest))
        
    elif problem_type == 'class':
        
        est = (est == est.max(axis=1, keepdims=True)).astype(int) - 1/Ytest.shape[1]
        TestErr = 100 - 100*npLA.norm(np.count_nonzero(est - Ytest, axis=1), 0)/len(Ytest)
    
    
    elapsed_time = time.time() - start_time
    
    return TestErr, elapsed_time



def RowNorm_KRR(s, q, m, lambda_, device_, problem_type, Xtrain, Ytrain, Xtest, Ytest):
    start_time = time.time()
    n = Xtrain.shape[0]
    d = Xtest.shape[1]
    
    rn_samplr = RNSampler(m)
    
    rn_samplr.Gen_Samples(s, q, torch.from_numpy(Xtrain))
    
    
    Features_train = rn_samplr.Apply_Sampler(torch.from_numpy(Xtrain)).data.numpy()
    Features_test = rn_samplr.Apply_Sampler(torch.from_numpy(Xtest)).data.numpy()
    
    
    w_star = npLA.solve(np.dot(Features_train.T, Features_train)+lambda_*np.eye(Features_train.shape[1]), np.dot(Features_train.T, Ytrain))
    
    est = np.dot(Features_test, w_star)
    
        
    if problem_type == 'regress':
        
        TestErr = npLA.norm(est - Ytest)/np.sqrt(len(Ytest))
        
    elif problem_type == 'class':
        
        est = (est == est.max(axis=1, keepdims=True)).astype(int) - 1/Ytest.shape[1]
        TestErr = 100 - 100*npLA.norm(np.count_nonzero(est - Ytest, axis=1), 0)/len(Ytest)
    
    
    elapsed_time = time.time() - start_time
    
    return TestErr, elapsed_time




def RowNorm_NTK_KRR(L, q, m, lambda_, device_, problem_type, Xtrain, Ytrain, Xtest, Ytest):
    start_time = time.time()
    n = Xtrain.shape[0]
    d = Xtest.shape[1]
    
    rn_samplr = NTK_RNSampler(m)
    
    rn_samplr.Gen_Samples(L, q, torch.from_numpy(Xtrain))
    
    
    Features_train = rn_samplr.Apply_Sampler(torch.from_numpy(Xtrain)).data.numpy()
    Features_test = rn_samplr.Apply_Sampler(torch.from_numpy(Xtest)).data.numpy()
    
    
    w_star = npLA.solve(np.dot(Features_train.T, Features_train)+lambda_*np.eye(Features_train.shape[1]), np.dot(Features_train.T, Ytrain))
    
    est = np.dot(Features_test, w_star)
    
        
    if problem_type == 'regress':
        
        TestErr = npLA.norm(est - Ytest)/np.sqrt(len(Ytest))
        
    elif problem_type == 'class':
        
        est = (est == est.max(axis=1, keepdims=True)).astype(int) - 1/Ytest.shape[1]
        TestErr = 100 - 100*npLA.norm(np.count_nonzero(est - Ytest, axis=1), 0)/len(Ytest)
    
    
    elapsed_time = time.time() - start_time
    
    return TestErr, elapsed_time






def RLS_KRR(s, q, m, lambda_, device_, problem_type, Xtrain, Ytrain, Xtest, Ytest):
    start_time = time.time()
    n = Xtrain.shape[0]
    d = Xtest.shape[1]
    
    rls_samplr = RLSSampler(m)
    
    rls_samplr.Gen_Samples(s, q, torch.from_numpy(Xtrain), lambda_)
    
    
    Features_train = rls_samplr.Apply_Sampler(torch.from_numpy(Xtrain)).data.numpy()
    Features_test = rls_samplr.Apply_Sampler(torch.from_numpy(Xtest)).data.numpy()
    
    
    w_star = npLA.solve(np.dot(Features_train.T, Features_train)+lambda_*np.eye(Features_train.shape[1]), np.dot(Features_train.T, Ytrain))
    
    est = np.dot(Features_test, w_star)
    
        
    if problem_type == 'regress':
        
        TestErr = npLA.norm(est - Ytest)/np.sqrt(len(Ytest))
        
    elif problem_type == 'class':
        
        est = (est == est.max(axis=1, keepdims=True)).astype(int) - 1/Ytest.shape[1]
        TestErr = 100 - 100*npLA.norm(np.count_nonzero(est - Ytest, axis=1), 0)/len(Ytest)
    
    
    elapsed_time = time.time() - start_time
    
    return TestErr, elapsed_time




def RLS_NTK_KRR(L, q, m, lambda_, device_, problem_type, Xtrain, Ytrain, Xtest, Ytest):
    start_time = time.time()
    n = Xtrain.shape[0]
    d = Xtest.shape[1]
    
    rls_samplr = NTK_RLSSampler(m)
    
    rls_samplr.Gen_Samples(L, q, torch.from_numpy(Xtrain), lambda_)
    
    
    Features_train = rls_samplr.Apply_Sampler(torch.from_numpy(Xtrain)).data.numpy()
    Features_test = rls_samplr.Apply_Sampler(torch.from_numpy(Xtest)).data.numpy()
    
    
    w_star = npLA.solve(np.dot(Features_train.T, Features_train)+lambda_*np.eye(Features_train.shape[1]), np.dot(Features_train.T, Ytrain))
    
    est = np.dot(Features_test, w_star)
    
        
    if problem_type == 'regress':
        
        TestErr = npLA.norm(est - Ytest)/np.sqrt(len(Ytest))
        
    elif problem_type == 'class':
        
        est = (est == est.max(axis=1, keepdims=True)).astype(int) - 1/Ytest.shape[1]
        TestErr = 100 - 100*npLA.norm(np.count_nonzero(est - Ytest, axis=1), 0)/len(Ytest)
    
    
    elapsed_time = time.time() - start_time
    
    return TestErr, elapsed_time




def RLS_old_KRR(s, q, m, lambda_, device_, problem_type, Xtrain, Ytrain, Xtest, Ytest):
    start_time = time.time()
    n = Xtrain.shape[0]
    d = Xtest.shape[1]
    
    rls_samplr = RLSSampler(m)
    
    rls_samplr.Gen_Samples_old_Version(s, q, torch.from_numpy(Xtrain), lambda_)
    
    
    Features_train = rls_samplr.Apply_Sampler(torch.from_numpy(Xtrain)).data.numpy()
    Features_test = rls_samplr.Apply_Sampler(torch.from_numpy(Xtest)).data.numpy()
    
    
    w_star = npLA.solve(np.dot(Features_train.T, Features_train)+lambda_*np.eye(Features_train.shape[1]), np.dot(Features_train.T, Ytrain))
    
    est = np.dot(Features_test, w_star)
    
        
    if problem_type == 'regress':
        
        TestErr = npLA.norm(est - Ytest)/np.sqrt(len(Ytest))
        
    elif problem_type == 'class':
        
        est = (est == est.max(axis=1, keepdims=True)).astype(int) - 1/Ytest.shape[1]
        TestErr = 100 - 100*npLA.norm(np.count_nonzero(est - Ytest, axis=1), 0)/len(Ytest)
    
    
    elapsed_time = time.time() - start_time
    
    return TestErr, elapsed_time





def FF_KRR(s, q, m, lambda_, device_, problem_type, Xtrain, Ytrain, Xtest, Ytest):
    start_time = time.time()
    n = Xtrain.shape[0]
    d = Xtest.shape[1]
        
    
    Features_train , Features_test = FF(s, m, torch.from_numpy(Xtrain), torch.from_numpy(Xtest))
    Features_train = Features_train.data.numpy()
    Features_test = Features_test.data.numpy()
    
    w_star = npLA.solve(np.dot(Features_train.T, Features_train)+lambda_*np.eye(Features_train.shape[1]), np.dot(Features_train.T, Ytrain))
    
    est = np.dot(Features_test, w_star)
    
        
    if problem_type == 'regress':
        
        TestErr = npLA.norm(est - Ytest)/np.sqrt(len(Ytest))
        
    elif problem_type == 'class':
        
        est = (est == est.max(axis=1, keepdims=True)).astype(int) - 1/Ytest.shape[1]
        TestErr = 100 - 100*npLA.norm(np.count_nonzero(est - Ytest, axis=1), 0)/len(Ytest)
    
    
    elapsed_time = time.time() - start_time
    
    return TestErr, elapsed_time



def Nystrom_KRR(s, q, m, lambda_, device_, problem_type, Xtrain, Ytrain, Xtest, Ytest):
    start_time = time.time()
    n = Xtrain.shape[0]
    d = Xtest.shape[1]
    
    Features_train, Features_test = NF(s, lambda_, m, Xtrain, Xtest)
    
    w_star = npLA.solve(np.dot(Features_train.T, Features_train)+lambda_*np.eye(Features_train.shape[1]), np.dot(Features_train.T, Ytrain))
    
    est = np.dot(Features_test, w_star)
    
        
    if problem_type == 'regress':
        
        TestErr = npLA.norm(est - Ytest)/np.sqrt(len(Ytest))
        
    elif problem_type == 'class':
        
        est = (est == est.max(axis=1, keepdims=True)).astype(int) - 1/Ytest.shape[1]
        TestErr = 100 - 100*npLA.norm(np.count_nonzero(est - Ytest, axis=1), 0)/len(Ytest)
    
    
    elapsed_time = time.time() - start_time
    
    return TestErr, elapsed_time




def NormDiff(s, q, m, lambda_, Xtrain, Xtest):
    start_time = time.time()
    n = Xtest.shape[0]
    d = Xtest.shape[1]
    
    rls_samplr = RLSSampler(m)
    
    rls_samplr.Gen_Samples(s, q, torch.from_numpy(Xtrain), lambda_)
    
    Features_test = rls_samplr.Apply_Sampler(torch.from_numpy(Xtest)).data.numpy()
    
    Kexact = np.ones((n,n))
    Gram = np.dot(Xtest, Xtest.T)
    
    for i in range(q):
        Kexact += ( ((2/s)**(i+1))/ math.factorial(i+1) ) * (Gram ** (i+1))
    
    Normalizer = np.exp(-(npLA.norm(Xtest, 2, axis=1)**2)/s)
    
    Kexact = (Kexact * Normalizer).T * Normalizer
    
    err = npLA.norm(np.dot(Features_test, Features_test.T) - Kexact, 'fro')/npLA.norm(Kexact, 'fro')
    print('elapsed time= ', time.time() - start_time)
    return err

def NormDiffRowNorm(s, q, m, lambda_, Xtrain, Xtest):
    
    n = Xtest.shape[0]
    d = Xtest.shape[1]
    
    rn_samplr = RNSampler(m)
    
    rn_samplr.Gen_Samples(s, q, torch.from_numpy(Xtrain))
    
    Features_test = rn_samplr.Apply_Sampler(torch.from_numpy(Xtest)).data.numpy()
    
    Kexact = np.ones((n,n))
    Gram = np.dot(Xtest, Xtest.T)
    
    for i in range(q):
        Kexact += ( ((2/s)**(i+1))/ math.factorial(i+1) ) * (Gram ** (i+1))
    
    Normalizer = np.exp(-(npLA.norm(Xtest, 2, axis=1)**2)/s)
    
    Kexact = (Kexact * Normalizer).T * Normalizer
    
    err = npLA.norm(np.dot(Features_test, Features_test.T) - Kexact, 'fro')/npLA.norm(Kexact, 'fro')
    
    return err



data = np.load('MNIST/mnistv.npz')

Xtrain = data['arr_0']
Ytrain = data['arr_1']
Xtest = data['arr_2']
Ytest = data['arr_3']

Xtest = Xtest - np.mean(Xtrain, axis=0)
Xtrain = Xtrain - np.mean(Xtrain, axis=0)


lambda_ = 0.1
m = 1000
L = 1
s = 1
q = 8
device_ = 'cpu'
problem_type = 'class'

TestErr, elapsed_time = FF_KRR(L, q, m, lambda_, device_, problem_type, Xtrain, Ytrain, Xtest, Ytest)

print('Elapsed Time=', elapsed_time, 'Error =', TestErr, 'lambda =', lambda_, 'Degree=', q, 'kernel depth=', L, 'number of features=', m)

# TestErr, elapsed_time = Nystrom_KRR(s, q, m, lambda_, device_, problem_type, Xtrain, Ytrain, Xtest, Ytest)

# print('Elapsed Time=', elapsed_time, 'Error =', TestErr, 'lambda =', lambda_, 'Degree=', q, 'kernel width=', s, 'number of features=', m)


# Err = NormDiff(s, q, m, lambda_, Xtrain, Xtest[0:2000,:])

# print('Error =', Err, 'Degree=', q, 'kernel width=', s, 'number of features=', m)
