#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Sep 30 11:48:05 2019

@author: zandieh
"""

import numpy as np
import torch
import torch.fft
from torch import linalg as LA
from scipy import linalg as sciLA
import math
from TensorSketchCuda import TensorSketch , SRHTCmplx_Stndrd
from RowNormSampling import RNSampler , NTK_RNSampler



class RLSSampler:
    
    def __init__(self, m):
            
        self.m = m
        
        self.S = [0 for i in range(m)]
        self.W = torch.zeros(m, dtype=torch.float64)
        
        self.D = torch.from_numpy(np.random.choice((-1,1), 2** int(np.log2(self.m) +1) ))
        self.P = torch.from_numpy(np.random.choice(len(self.D), int(self.m/8)))

        
    
    def Gen_Samples(self, s, q, X, lambda_):
        
        self.n = X.shape[0]
        self.d = X.shape[1]
        self.s = s
        self.lambda_ = 8*lambda_
        self.q = q
        
        rownormSampler = RNSampler(len(self.D))
        
        rownormSampler.Gen_Samples(s, q, X)
        
        B = rownormSampler.Apply_Sampler(X)
        B = SRHTCmplx_Stndrd(B, self.P, self.D)
        B = torch.cat((B.real, B.imag), 1)
        
        TN_DS = TensorSketch(self.d, 2** int(np.log2(self.m) - 2), self.q, True, 'cpu')
        TN_DS.Sketch_Kernel_Inv(B, self.lambda_)
        TN = TN_DS.Sketch_pre_SRHT(X)
        
        Exp = np.exp(-(LA.norm(X, dim=1)**2)/s)
        
                
        coeff_stablizer = min(1, LA.norm(Exp))
        Exp = Exp / coeff_stablizer
        Exp = Exp.type(TN[0].dtype)
        
        
        Y = X.type(TN[0].dtype)
        
        f = np.zeros(q+1)
        f[0] = LA.norm(torch.matmul( TN[q-1].T , Exp) )**2
        for i in range(q):
            f[i+1] = LA.norm( torch.mm( np.sqrt(((2/s)**(i+1))/math.factorial(i+1)) *  TN[q-1-i].T * Exp , Y) )**2
        f = f/np.sum(f)
        
        degs = np.random.choice(q+1, self.m, p=f)
        
        u, counts = np.unique(degs, return_counts=True)
        deg_counts = np.zeros(q+1, dtype=int)
        deg_counts[u] = counts
        print(deg_counts)
        deg_revcumsum = np.append( np.cumsum(deg_counts[::-1])[::-1], 0)
        deg_cumsum = np.cumsum(deg_counts)
        for i in range(deg_cumsum[0]):
            self.S[i] = torch.zeros(1, dtype=torch.int)
            self.W[i] = np.sqrt(1/(self.m * f[0]))
        
        
        beta_ = self.m * torch.ones(deg_revcumsum[1])
        D = (torch.ones(deg_revcumsum[2], self.n) * Exp).T
        for b in range(1, q+1):
            p_1 = LA.norm( torch.mm( np.sqrt(((2/s)**(b))/math.factorial(b)) *  TN[q-b].T * Exp , Y) , dim=0)**2
            p_1 = p_1 / torch.sum(p_1)                
            p_1 = p_1.data.numpy()
            I_1 = np.random.choice(self.d, deg_counts[b], p = p_1 )
            
            beta_[deg_revcumsum[b+1]:deg_revcumsum[b]] = beta_[deg_revcumsum[b+1]:deg_revcumsum[b]] * p_1[I_1]
            if b >1:
                D[:, deg_revcumsum[b+1]:deg_revcumsum[b]] = D[:, deg_revcumsum[b+1]:deg_revcumsum[b]] * Y[:,I_1]
                for j, i in enumerate(range(deg_cumsum[b-1], deg_cumsum[b])):
                    self.S[i] = torch.zeros(b+1, dtype=torch.int)
                    self.S[i][0:2] = torch.tensor([b,I_1[j]])
            else:
                beta_[deg_revcumsum[b+1]:deg_revcumsum[b]] *= f[b]
                for j, i in enumerate(range(deg_cumsum[b-1], deg_cumsum[b])):
                    self.S[i] = torch.tensor([b,I_1[j]], dtype=torch.int)
                    self.W[i] = np.sqrt(1/beta_[j+deg_revcumsum[b+1]])


        for a in range(2,q+1):
            for b in range(a, q+1):
                p_a = LA.norm( torch.matmul( torch.permute( np.sqrt(((2/s)**(b-a))/math.factorial(b-a)) *  TN[q+a-1-b].T.unsqueeze(2).repeat(1,1, deg_counts[b]) * D[:, deg_revcumsum[b+1]:deg_revcumsum[b]] , (2,0,1) ) , Y) , dim=1)**2
                p_a = p_a / torch.sum(p_a, dim=1).unsqueeze(-1)            
                p_a = p_a.data.numpy()
                I_a = np.zeros(deg_counts[b], dtype=int)
                for i in range(deg_counts[b]):
                    I_a[i] = np.random.choice(self.d, 1, p = p_a[i,:] )[0]
                
                beta_[deg_revcumsum[b+1]:deg_revcumsum[b]] = beta_[deg_revcumsum[b+1]:deg_revcumsum[b]] * p_a[np.arange(deg_counts[b]), I_a]
                if b >a:
                    D[:, deg_revcumsum[b+1]:deg_revcumsum[b]] = D[:, deg_revcumsum[b+1]:deg_revcumsum[b]] * Y[:,I_a]
                    for j, i in enumerate(range(deg_cumsum[b-1], deg_cumsum[b])):
                        self.S[i][a] = I_a[j]
                else:
                    beta_[deg_revcumsum[b+1]:deg_revcumsum[b]] *= f[b]
                    for j, i in enumerate(range(deg_cumsum[b-1], deg_cumsum[b])):
                        self.S[i][a] = I_a[j]
                        self.W[i] = np.sqrt(1/beta_[j+deg_revcumsum[b+1]])


        
        
        
        # for l, b in enumerate(degs):
            
        #     beta_ = self.m * f[b]
        #     D = Exp
            
        #     self.S[l] = torch.zeros(b+1, dtype=torch.int)
        #     self.S[l][0] = b
            
        #     for a in range(b):
        #         p_a = LA.norm( torch.mm( TN[q+a-b].T * D , Y) , dim=0)**2
                
        #         p_a = p_a / torch.sum(p_a)                
        #         p_a = p_a.data.numpy()
                
        #         i_a = np.random.choice(self.d, 1, p = p_a )[0]
        #         self.S[l][a+1] = i_a
                
        #         beta_ = beta_ * p_a[i_a]
        #         D = D * Y[:,i_a]
            
            
        #     self.W[l] = np.sqrt(1/beta_)
            


    def Apply_Sampler(self, X):
        
        n = X.shape[0]
                
        Z = (torch.ones(self.m, n) * np.exp(-(LA.norm(X, dim=1)**2)/self.s)).T * self.W
        
        for i in range(self.m):
            Z[:,i] = np.sqrt(((2/self.s)**self.S[i][0])/math.factorial(self.S[i][0])) * Z[:,i]
            
            for a in range(self.S[i][0]):
                
                Z[:,i] = Z[:,i] * X[:, self.S[i][1+a]]
        
        return Z





class NTK_RLSSampler:
    
    def __init__(self, m):
            
        self.m = m
        
        self.S = [0 for i in range(m)]
        self.W = torch.zeros(m, dtype=torch.float64)
        
        self.D = torch.from_numpy(np.random.choice((-1,1), 2** int(np.log2(self.m) +1) ))
        self.P = torch.from_numpy(np.random.choice(len(self.D), int(self.m/8)))

        
    
    def Gen_Samples(self, L, q, X, lambda_):
        
        self.n = X.shape[0]
        self.d = X.shape[1]
        self.L = L
        self.lambda_ = 8*lambda_
        self.q = q
        
        rownormSampler = NTK_RNSampler(len(self.D))
        
        rownormSampler.Gen_Samples(L, q, X)
        
        B = rownormSampler.Apply_Sampler(X)
        B = SRHTCmplx_Stndrd(B, self.P, self.D)
        B = torch.cat((B.real, B.imag), 1)
        
        
        norm_X = LA.norm(X, dim=1)
        Normalizer = torch.where(norm_X>0, norm_X, 1.0)
        Xnormalized = ((X.T / Normalizer).T)

        TN_DS = TensorSketch(self.d, 2** int(np.log2(self.m) - 2), self.q, True, 'cpu')
        TN_DS.Sketch_Kernel_Inv(B, self.lambda_)
        TN = TN_DS.Sketch_pre_SRHT(Xnormalized)
        
        self.NTK_coeff = rownormSampler.NTK_coeff
        
        Normalizer = Normalizer.type(TN[0].dtype)
        
        Y = Xnormalized.type(TN[0].dtype)
        
        f = np.zeros(q+1)
        f[0] = self.NTK_coeff[0] * LA.norm(torch.matmul( TN[q-1].T , Normalizer) )**2
        for i in range(q):
            f[i+1] = self.NTK_coeff[i+1] * LA.norm( torch.mm( TN[q-1-i].T * Normalizer , Y) )**2
        f = f/np.sum(f)
        
        degs = np.random.choice(q+1, self.m, p=f)
        
        u, counts = np.unique(degs, return_counts=True)
        deg_counts = np.zeros(q+1, dtype=int)
        deg_counts[u] = counts
        print(deg_counts)
        deg_revcumsum = np.append( np.cumsum(deg_counts[::-1])[::-1], 0)
        deg_cumsum = np.cumsum(deg_counts)
        for i in range(deg_cumsum[0]):
            self.S[i] = torch.zeros(1, dtype=torch.int)
            self.W[i] = np.sqrt(1/(self.m * f[0]))
        
        
        beta_ = self.m * torch.ones(deg_revcumsum[1])
        D = (torch.ones(deg_revcumsum[2], self.n) * Normalizer).T
        for b in range(1, q+1):
            p_1 = LA.norm( torch.mm( TN[q-b].T * Normalizer , Y) , dim=0)**2
            p_1 = p_1 / torch.sum(p_1)                
            p_1 = p_1.data.numpy()
            I_1 = np.random.choice(self.d, deg_counts[b], p = p_1 )
            
            beta_[deg_revcumsum[b+1]:deg_revcumsum[b]] = beta_[deg_revcumsum[b+1]:deg_revcumsum[b]] * p_1[I_1]
            if b >1:
                D[:, deg_revcumsum[b+1]:deg_revcumsum[b]] = D[:, deg_revcumsum[b+1]:deg_revcumsum[b]] * Y[:,I_1]
                for j, i in enumerate(range(deg_cumsum[b-1], deg_cumsum[b])):
                    self.S[i] = torch.zeros(b+1, dtype=torch.int)
                    self.S[i][0:2] = torch.tensor([b,I_1[j]])
            else:
                beta_[deg_revcumsum[b+1]:deg_revcumsum[b]] *= f[b]
                for j, i in enumerate(range(deg_cumsum[b-1], deg_cumsum[b])):
                    self.S[i] = torch.tensor([b,I_1[j]], dtype=torch.int)
                    self.W[i] = np.sqrt(1/beta_[j+deg_revcumsum[b+1]])


        for a in range(2,q+1):
            for b in range(a, q+1):
                p_a = LA.norm( torch.matmul( torch.permute( TN[q+a-1-b].T.unsqueeze(2).repeat(1,1, deg_counts[b]) * D[:, deg_revcumsum[b+1]:deg_revcumsum[b]] , (2,0,1) ) , Y) , dim=1)**2
                p_a = p_a / torch.sum(p_a, dim=1).unsqueeze(-1)            
                p_a = p_a.data.numpy()
                I_a = np.zeros(deg_counts[b], dtype=int)
                for i in range(deg_counts[b]):
                    I_a[i] = np.random.choice(self.d, 1, p = p_a[i,:] )[0]
                
                beta_[deg_revcumsum[b+1]:deg_revcumsum[b]] = beta_[deg_revcumsum[b+1]:deg_revcumsum[b]] * p_a[np.arange(deg_counts[b]), I_a]
                if b >a:
                    D[:, deg_revcumsum[b+1]:deg_revcumsum[b]] = D[:, deg_revcumsum[b+1]:deg_revcumsum[b]] * Y[:,I_a]
                    for j, i in enumerate(range(deg_cumsum[b-1], deg_cumsum[b])):
                        self.S[i][a] = I_a[j]
                else:
                    beta_[deg_revcumsum[b+1]:deg_revcumsum[b]] *= f[b]
                    for j, i in enumerate(range(deg_cumsum[b-1], deg_cumsum[b])):
                        self.S[i][a] = I_a[j]
                        self.W[i] = np.sqrt(1/beta_[j+deg_revcumsum[b+1]])


        
            


    def Apply_Sampler(self, X):
        
        n = X.shape[0]
        norm_X = LA.norm(X, dim=1)
        Normalizer = torch.where(norm_X>0, norm_X, 1.0)
        Xnormalized = ((X.T / Normalizer).T)

        Z = (torch.ones(self.m, n) * norm_X).T * self.W
        
        for i in range(self.m):
            Z[:,i] = np.sqrt(self.NTK_coeff[self.S[i][0]]) * Z[:,i]
            
            for a in range(self.S[i][0]):
                
                Z[:,i] = Z[:,i] * Xnormalized[:, self.S[i][1+a]]
        
        return Z



    

    def Gen_Samples_old_Version(self, s, q, X, lambda_):
        
        self.n = X.shape[0]
        self.d = X.shape[1]
        self.s = s
        self.lambda_ = 8*lambda_
        self.q = q
        
        rownormSampler = RNSampler(len(self.D))
        
        rownormSampler.Gen_Samples(s, q, X)
        
        B = rownormSampler.Apply_Sampler(X)
        B = SRHTCmplx_Stndrd(B, self.P, self.D)
        B = torch.cat((B.real, B.imag), 1)

        TN_DS = TensorSketch(self.d, 2** int(np.log2(self.m) - 2), self.q, False, 'cpu')
        U = TN_DS.Sketch(X)
        
        H = torch.from_numpy( np.sqrt(1/TN_DS.m) * np.random.randn(self.n, TN_DS.m) )
        M = np.sqrt(1/self.lambda_) * H
        
        u, sigm, vh = sciLA.svd(B.to('cpu').data.numpy(), full_matrices=False)
        sigm = 1/(np.sqrt(sigm**2+self.lambda_)) - np.sqrt(1/self.lambda_)
        
        u = torch.from_numpy( u )
        
        M = M + torch.mm(u*sigm, torch.mm(u.T, H))

        
        m_prime = U[0].shape[1]
        
        
        
        Exp = np.exp(-(LA.norm(X, dim=1)**2)/s)
        
        coeff_stablizer = min(1, LA.norm(Exp))
        Exp = Exp / coeff_stablizer
        Exp = Exp.type(U[0].dtype)

        
        M = M.type(U[0].dtype)
        
        R = torch.ones((len(TN_DS.D)+1, self.n), dtype=U[0].dtype)
        for i in range(q):
            R[m_prime*i+1:m_prime*(i+1)+1,:] = np.sqrt(((2/s)**(i+1))/ math.factorial(i+1)) * U[q-i-1].T
        R = R*Exp
        RM = torch.matmul(R, M)
        
        
        p = LA.norm(RM, dim=0)**2
        p = p / torch.sum(p)
        p = p.data.numpy()
        
        J = np.random.choice(len(p), self.m, p=p)
        U.append(torch.ones((self.n,1), dtype=U[0].dtype))
        
        
        Y = X.type(U[0].dtype)
        
        
        y = np.zeros((q+1,len(p)))
        for j in range(len(p)):
            y[0,j] = RM[0,j]**2
            for i in range(1,q+1):
                y[i,j] = LA.norm(RM[(i-1)*m_prime+1:i*m_prime+1,j])**2
            y[:,j] = y[:,j]/np.sum(y[:,j])

        
        for l, j in enumerate(J):
        
            w = np.random.choice(q+1, 1, p=y[:,j])[0]
            I = np.ones(w)
            
            D = Exp * M[:,j]
            
            for a in range(w):
                pa = LA.norm(torch.matmul(Y.T*D, U[a+q+1-w]), dim=1)**2
                pa = pa/torch.sum(pa)
                pa = pa.data.numpy()
                I[a] = np.random.choice(self.d, 1, p=pa)[0]
                D = D * Y[:,int(I[a])]

        
            beta_ = np.sqrt(((2/s)**(w))/ math.factorial(w)) * Exp
            for a in range(w):
                beta_ = beta_ * Y[:,int(I[a])]
            self.W[l] = np.sqrt((LA.norm(RM)**2)/(self.m*LA.norm(torch.matmul(beta_, M))**2))
            
            self.S[l] = np.concatenate((np.array([w]), I)).astype(int)
        

