#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jan 23 16:16:22 2022

@author: amir
"""
import numpy as np
import torch
import torch.fft
from torch import linalg as LA
import math
import quadprog



def quadprog_solve_qp(P, q, G=None, h=None, A=None, b=None):
    qp_G = .5 * (P + P.T +0.00001*np.eye(P.shape[0]))   # make sure P is symmetric
    qp_a = -q
    if A is not None:
        qp_C = -np.vstack([A, G]).T
        qp_b = -np.hstack([b, h])
        meq = A.shape[0]
    else:  # no equality constraint
        qp_C = -G.T
        qp_b = -h
        meq = 0
    return quadprog.solve_qp(qp_G, qp_a, qp_C, qp_b, meq)[0]

def NTK(L,q):
    n=15*L+5*q
    Y = np.zeros((201+n,L+1))
    Y[:,0] = np.sort(np.concatenate((np.linspace(-1.0, 1.0, num=201), np.cos((2*np.arange(n)+1)*np.pi / (4*n))), axis=0))

    m_ = Y.shape[0]

    for i in range(L):
        Y[:,i+1] = (np.sqrt(1-Y[:,i]**2) + Y[:,i]*(np.pi - np.arccos(Y[:,i])))/np.pi
    
    y = np.zeros(m_)
    for i in range(L+1):
        z = Y[:,i]
        for j in range(i,L):
            z = z*(np.pi - np.arccos(Y[:,j]))/np.pi
        y = y + z
    
    Z = np.zeros((m_,q+1))
    Z[:,0] = np.ones(m_)
    for i in range(q):
        Z[:,i+1] = Z[:,i] * Y[:,0]
    
    
    weight_ = np.linspace(0.0, 1.0, num=m_) + 2/L
    w = y * weight_
    U = Z.T * weight_

    beta_ = quadprog_solve_qp(np.dot(U, U.T), -np.dot(U,w) , np.concatenate((Z[0:m_-1,:]-Z[1:m_,:], -np.eye(q+1)),axis=0), np.zeros(q+m_))
    beta_[beta_ < 0.00001] = 0
    
    return beta_





def OblvPolySNTKFeat(L, tensr_sktch, X):
    q = tensr_sktch.q
    n = X.shape[0]
    
    NTK_coeff = NTK(L, q)
    
    norm_X = LA.norm(X, dim=1)
    Normalizer = torch.where(norm_X>0, norm_X, 1.0)
    Xnormalized = ((X.T / Normalizer).T)
    
    U = tensr_sktch.Sketch(Xnormalized)
    
    m = U[0].shape[1]
    
    Z = torch.zeros((len(tensr_sktch.D),n), dtype=torch.cfloat, device=tensr_sktch.device_)
    for i in range(q):
        Z[m*i:m*(i+1)] = np.sqrt( NTK_coeff[i+1] ) * U[q-i-1].T
        U[q-i-1]=0
    
    Z = (np.sqrt(1/len(tensr_sktch.P))*torch.fft.fftn(Z.T*tensr_sktch.D, dim=1)[:,tensr_sktch.P])
    Z = (Z.T * Normalizer).T
    
    return torch.cat(( np.sqrt(NTK_coeff[0]) * Normalizer.reshape((n,1)), torch.cat((Z.real, Z.imag), 1)), 1)
    


def OblvPolySFeat(s, tensr_sktch, X):
    q = tensr_sktch.q
    n = X.shape[0]

    Normalizer = np.exp(-(LA.norm(X, dim=1)**2)/s) 
    
    U = tensr_sktch.Sketch(X)
    
    m = U[0].shape[1]
    
    Z = torch.zeros((len(tensr_sktch.D),n), dtype=torch.cfloat, device=tensr_sktch.device_)
    for i in range(q):
        Z[m*i:m*(i+1)] = np.sqrt( ((2/s)**(i+1))/ math.factorial(i+1) ) * U[q-i-1].T
        U[q-i-1]=0
    
    Z = (np.sqrt(1/len(tensr_sktch.P))*torch.fft.fftn(Z.T*tensr_sktch.D, dim=1)[:,tensr_sktch.P])
    Z = (Z.T * Normalizer).T
    
    return torch.cat((Normalizer.reshape((n,1)), torch.cat((Z.real, Z.imag), 1)), 1)
    
