#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Sep 30 11:48:05 2019

@author: zandieh
"""
import numpy as np
import torch
import torch.fft
from scipy import linalg as LA



def TSRHTCmplx(X1, X2, P, D):
    
    Xhat1 = torch.fft.fftn(X1 * D[0,:], dim=1)[:,P[0,:]]
    Xhat2 = torch.fft.fftn(X2 * D[1,:], dim=1)[:,P[1,:]]
    
    Y = np.sqrt(1/P.shape[1])*(Xhat1 * Xhat2)
    
    return Y
    
def SRHTCmplx_Stndrd(X, P, D):
    return np.sqrt(1/len(P)) * torch.fft.fftn(X * D, dim=1)[:,P]

class TensorSketch:
    def __init__(self, d, m, q, presketch_srht, dev):
        
        
        self.m = m
        self.q = q
        self.device_ = dev
    
        self.Tree_D = [0 for i in range((self.q-1).bit_length())]
        self.Tree_P = [0 for i in range((self.q-1).bit_length())]
        
        self.presketch_srht = presketch_srht
        
        m_=int(self.m/4)
        q_ = int(self.q/2)
        
        if presketch_srht:
            
            self.Pre_SRHT_D = torch.from_numpy(np.random.choice((-1,1), d)).to(self.device_)
            self.Pre_SRHT_P = torch.from_numpy(np.random.choice(d, (q_,2, 2*m_))).to(self.device_)
            
            self.d = 2*m_
            
        else:
            
            self.d = d
    
        
        for i in range((self.q-1).bit_length()):
            if i == 0:
                self.Tree_P[i] = torch.from_numpy(np.random.choice(self.d, (q_,2,m_))).to(self.device_)
                self.Tree_D[i] = torch.from_numpy(np.random.choice((-1,1), (q_,2,self.d))).to(self.device_)
            else:
                self.Tree_P[i] = torch.from_numpy(np.random.choice(m_, (q_,2,m_))).to(self.device_)
                self.Tree_D[i] = torch.from_numpy(np.random.choice((-1,1), (q_,2,m_))).to(self.device_)
            q_ = int(q_/2)
        
        self.D = torch.from_numpy(np.random.choice((-1,1), self.q*m_)).to(self.device_)
        self.P = torch.from_numpy(np.random.choice(self.q*m_, int(self.m/2-1))).to(self.device_)
        
        
            
        

    def Sketch(self, X):
        n=X.shape[0]
        lgq = len(self.Tree_D)
        V = [0 for i in range(lgq)]
        #E1 = torch.cat((torch.ones((n,1), device=self.device_), torch.zeros((n,X.shape[1]-1), device=self.device_)), 1)
        
        for i in range(lgq):
            q = self.Tree_D[i].shape[0]
            V[i] = torch.zeros((q,n,self.Tree_P[i].shape[2]), dtype=torch.cfloat, device=self.device_)
            for j in range(q):
                if i == 0:
                    V[i][j,:,:] = TSRHTCmplx(X, X, self.Tree_P[i][j,:,:], self.Tree_D[i][j,:,:])
                else:
                    V[i][j,:,:] = TSRHTCmplx(V[i-1][2*j,:,:], V[i-1][2*j+1,:,:], self.Tree_P[i][j,:,:], self.Tree_D[i][j,:,:])
    
        U = [0 for i in range(2**lgq)]
        U[0] = V[lgq-1][0,:,:].detach().clone()
        
        SetE1 = {}
        
        for j in range(1,len(U)):
            p = int((j-1)/2)
            for i in range(lgq):
                if j%(2**(i+1)) == 0 :
                    SetE1[(i,p)] = 1
                    #V[i][p,:,:] = torch.cat((torch.ones((n,1)), torch.zeros((n,V[i].shape[2]-1))), 1)
                else:
                    if i == 0:
                        V[i][p,:,:] = SRHTCmplx_Stndrd(X, self.Tree_P[i][p,0,:], self.Tree_D[i][p,0,:])
                    else:
                        if (i-1,2*p) in SetE1:
                            V[i][p,:,:] = V[i-1][2*p+1,:,:]
                        else:
                            V[i][p,:,:] = TSRHTCmplx(V[i-1][2*p,:,:], V[i-1][2*p+1,:,:], self.Tree_P[i][p,:,:], self.Tree_D[i][p,:,:])
                p = int(p/2)
            U[j] = V[lgq-1][0,:,:].detach().clone()
        
        return U
    
    
    def Sketch_Kernel_Inv(self, B, lambda_):
        
        n = B.shape[0]
        
        # # ################
        # H = torch.from_numpy( np.sqrt(1/self.Pre_SRHT_P.shape[2]) * np.random.randn(n, self.Pre_SRHT_P.shape[2]) ).to(self.device_)
        # self.M = H

        if self.Pre_SRHT_P.shape[2] < n:
            H = torch.from_numpy( np.sqrt(1/self.Pre_SRHT_P.shape[2]) * np.random.randn(n, self.Pre_SRHT_P.shape[2]) ).to(self.device_)
            M = np.sqrt(1/lambda_) * H
            
            u, sigm, vh = LA.svd(B.to('cpu').data.numpy(), full_matrices=False)
            sigm = 1/(np.sqrt(sigm**2+lambda_)) - np.sqrt(1/lambda_)
            
            u = torch.from_numpy( u ).to(self.device_)
            
            self.M = M + torch.mm(u*sigm, torch.mm(u.T, H))
        
        else:
            
            M = torch.from_numpy( LA.inv(torch.mm(B, B.T).to('cpu').data.numpy() + lambda_*np.eye(n) ) ).to(self.device_)
            self.M = torch.cat( (M , torch.zeros((n, self.Pre_SRHT_P.shape[2] - n), device=self.device_) ) , 1)
            

    def Sketch_pre_SRHT(self, X):
        
        n=X.shape[0]
        lgq = len(self.Tree_D)
        V = [0 for i in range(lgq)]
        
        Y = np.sqrt(1/self.Pre_SRHT_P.shape[2])* torch.fft.fftn(X * self.Pre_SRHT_D, dim=1)
        #E1 = torch.cat((torch.ones((n,1), device=self.device_), torch.zeros((n,X.shape[1]-1), device=self.device_)), 1)
        
        for i in range(lgq):
            q = self.Tree_D[i].shape[0]
            V[i] = torch.zeros((q,n,self.Tree_P[i].shape[2]), dtype=torch.cfloat, device=self.device_)
            for j in range(q):
                if i == 0:
                    if j==q-1:
                        V[i][j,:,:] = TSRHTCmplx( Y[:,self.Pre_SRHT_P[j,0,:]] , self.M , self.Tree_P[i][j,:,:], self.Tree_D[i][j,:,:])
                    else:
                        V[i][j,:,:] = TSRHTCmplx( Y[:,self.Pre_SRHT_P[j,0,:]] , Y[:,self.Pre_SRHT_P[j,1,:]] , self.Tree_P[i][j,:,:], self.Tree_D[i][j,:,:])
                else:
                    V[i][j,:,:] = TSRHTCmplx(V[i-1][2*j,:,:], V[i-1][2*j+1,:,:], self.Tree_P[i][j,:,:], self.Tree_D[i][j,:,:])
    
        U = [0 for i in range(2**lgq)]
        U[0] = np.sqrt(2)*V[lgq-1][0,:,0:int(V[lgq-1].shape[2]/2) ].detach().clone()
        U[0] = torch.cat((U[0].real, U[0].imag), 1)
        
        SetE1 = {}
        
        for j in range(1,len(U)):
            p = int((j-1)/2)
            for i in range(lgq):
                if j%(2**(i+1)) == 0 :
                    SetE1[(i,p)] = 1
                    #V[i][p,:,:] = torch.cat((torch.ones((n,1)), torch.zeros((n,V[i].shape[2]-1))), 1)
                else:
                    if i == 0:
                        if j == len(U)-1:
                            V[i][p,:,:] = SRHTCmplx_Stndrd(self.M, self.Tree_P[i][p,0,:], self.Tree_D[i][p,0,:])
                        else:
                            V[i][p,:,:] = np.sqrt(self.Pre_SRHT_P.shape[2]/self.Tree_P[i].shape[2])* Y[:,self.Pre_SRHT_P[p,0,0:self.Tree_P[i].shape[2]]].detach().clone()
                    else:
                        if (i-1,2*p) in SetE1:
                            V[i][p,:,:] = V[i-1][2*p+1,:,:]
                        else:
                            V[i][p,:,:] = TSRHTCmplx(V[i-1][2*p,:,:], V[i-1][2*p+1,:,:], self.Tree_P[i][p,:,:], self.Tree_D[i][p,:,:])
                p = int(p/2)
            U[j] = np.sqrt(2)*V[lgq-1][0,:,0:int(V[lgq-1].shape[2]/2) ].detach().clone()
            
            U[j] = torch.cat((U[j].real, U[j].imag), 1)
            
        return U



